/******************************************************************
 MIT License http://www.opensource.org/licenses/mit-license.php
 Author Mora <qiuzhongleiabc@126.com> (https://github.com/qiu8310)
*******************************************************************/

import { TagItem } from '.'
import { CustomOptions, getCustomComponents } from './custom'
import {
  apicomponents,
  Component,
  ComponentAttr,
  CustomAttr,
  getComponentAttrMarkdown,
  getComponentMarkdown,
  LanguageConfig,
} from './dev/'

interface TagAttrItem {
  attr: CustomAttr
  markdown: string
}

// 优先级
// customs > natives > basics

/**
 * 自动补全支持的所有的 tag
 * @param {CustomOptions} co 用于解析自定义组件的配置
 */
export async function autoApiName(lc: LanguageConfig, co?: CustomOptions) {
  let natives: TagItem[] = [...lc.components, ...apicomponents].map(mapComponent)
  let customs: TagItem[] = (await getCustomComponents(co)).map(mapComponent)
  return {
    customs,
    natives,
  }
}
function mapComponent(component: Component) {
  return { component, markdown: getComponentMarkdown(component) }
}
/**
 * 自动补全 Api 上的属性
 *
 * @export
 * @param {string} tagName 当前 tag 的名称
 * @param {{[key: string]: string}} attrs 当前已经写了的属性的集合
 * @param {CustomOptions} co 用于解析自定义组件的配置
 */
export async function autoApiAttr(
  tagName: string,
  tagAttrs: { [key: string]: string | boolean },
  lc: LanguageConfig,
  co?: CustomOptions
) {
  let attrs = await getApiAttrs(tagName, tagAttrs, lc, co)

  // 属性不能是已经存在的，也不能是事件
  let filter = createComponentFilter(tagAttrs, false)

  return {
    natives: attrs.filter(filter).map(mapComponentAttr) as TagAttrItem[],
  }
}

async function getApiAttrs(
  tagName: string,
  tagAttrs: { [key: string]: string | boolean },
  lc: LanguageConfig,
  co?: CustomOptions
) {
  let comp = await getComponent(tagName, lc, co)
  return comp ? getAvailableAttrsFromComponent(comp, tagAttrs) : []
}

function mapComponentAttr(attr: ComponentAttr) {
  return { attr, markdown: getComponentAttrMarkdown(attr) } as TagAttrItem
}

async function getComponent(tagName: string, lc: LanguageConfig, co?: CustomOptions) {
  let comp = [...lc.components, ...apicomponents].find(c => c.name === tagName)
  if (!comp) {
    comp = (await getCustomComponents(co)).find(c => c.name === tagName)
  }
  return comp
}

function getAvailableAttrsFromComponent(
  comp: Component,
  tagAttrs: { [key: string]: string | boolean }
): ComponentAttr[] {
  let attrs = comp.attrs || []
  let results = attrs.filter(a => tagAttrs[a.name] == null) // 先取出没有写的属性
    // 如果没写的属性中有 subAttrs，则要把它们全取出来
  ;[...results].forEach(a => {
    if (a.subAttrs) {
      a.subAttrs.forEach(s => {
        s.attrs.forEach(suba => {
          if (results.every(_ => _.name !== suba.name)) results.push(suba) // 去重
        })
      })
    }
  })

  // 写了的属性需要找出 subAttrs
  attrs
    .filter(a => tagAttrs[a.name] != null && a.subAttrs && a.subAttrs.length)
    .forEach(a => {
      let sub = (a.subAttrs || []).find(s => s.equal === tagAttrs[a.name])
      if (sub) results.push(...sub.attrs.filter(sa => tagAttrs[sa.name] == null))
    })
  return results
}

function createComponentFilter(existsTagAttrs: { [key: string]: string | boolean }, event?: boolean) {
  return (attr: ComponentAttr) => {
    // let isEvent = attr.name.startsWith('bind') || attr.name.startsWith('catch')
    let isEvent = false
    return existsTagAttrs[attr.name] == null && (event == null || (event ? isEvent : !isEvent))
  }
}
